@@ -11,29 +11,40 @@
 
 import copy
 import functools
+import inspect
 import itertools
 import warnings
 
 from abc import ABC
 from collections import OrderedDict
-from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
 import numpy as np
 
 from matplotlib.axes import Axes
 from matplotlib.figure import Figure
 from numpy import ndarray
+from qutip import Qobj
 
 import scqubits.core.central_dispatch as dispatch
 import scqubits.core.descriptors as descriptors
 import scqubits.core.sweeps as sweeps
 import scqubits.io_utils.fileio_serializers as serializers
-import scqubits.settings as settings
 import scqubits.utils.cpu_switch as cpu_switch
 import scqubits.utils.misc as utils
 import scqubits.utils.plotting as plot
 
-from scqubits.core._param_sweep import _ParameterSweep
+from scqubits import settings as settings
 from scqubits.core.hilbert_space import HilbertSpace
 from scqubits.core.namedslots_array import (
     NamedSlotsNdarray,
@@ -42,11 +53,12 @@ from scqubits.core.namedslots_array import (
 )
 from scqubits.core.oscillator import Oscillator
 from scqubits.core.qubit_base import QubitBaseClass
-from scqubits.core.spectrum_lookup import SpectrumLookupMixin
+from scqubits.core.spec_lookup import SpectrumLookupMixin
 from scqubits.core.storage import SpectrumData
 
 if TYPE_CHECKING:
     from scqubits.io_utils.fileio import IOData
+    from scqubits.legacy._param_sweep import _ParameterSweep
 
 if settings.IN_IPYTHON:
     from tqdm.notebook import tqdm
@@ -68,17 +80,21 @@ class ParameterSweepBase(ABC):
     StoredSweep
     """
 
-    _parameters = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
-    _evals_count = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
-    _data = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _parameters: Parameters = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _evals_count: int = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _data: Dict[str, Any] = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
     _hilbertspace: HilbertSpace
 
     _out_of_sync = False
-    _current_param_indices: Union[NpIndices, None]
+    _current_param_indices: NpIndices
 
     def get_subsys(self, index: int) -> QuantumSys:
         return self._hilbertspace[index]
 
+    @property
+    def hilbertspace(self):
+        return self._hilbertspace
+
     def get_subsys_index(self, subsys: QuantumSys) -> int:
         return self._hilbertspace.get_subsys_index(subsys)
 
@@ -103,12 +119,15 @@ class ParameterSweepBase(ABC):
         # <Sweep>[p1, p2, ...].dressed_eigenstates()
         if isinstance(key, tuple):
             self._current_param_indices = convert_to_std_npindex(key, self._parameters)
-        elif isinstance(key, (int, slice)):
+        elif isinstance(key, slice):
             if key == slice(None) or key == slice(None, None, None):
                 key = (key,) * len(self._parameters)
             else:
                 key = (key,)
             self._current_param_indices = convert_to_std_npindex(key, self._parameters)
+        elif isinstance(key, np.integer):
+            key = (key,)
+            self._current_param_indices = convert_to_std_npindex(key, self._parameters)
         return self
 
     def receive(self, event: str, sender: object, **kwargs) -> None:
@@ -131,6 +150,21 @@ class ParameterSweepBase(ABC):
             elif event == "PARAMETERSWEEP_UPDATE" and sender is self:
                 self._out_of_sync = True
 
+    def set_update_func(self, update_hilbertspace: Callable) -> Callable:
+        """Account for the two possible signatures of the `update_hilbertspace`
+        function. Inspect whether a `self` argument is given. If not, return a
+        function that accepts `self` as a dummy argument."""
+        arguments = inspect.signature(update_hilbertspace)
+        if len(arguments.parameters) == len(self._parameters) + 1:
+            # update_hilbertspace function already includes self argument
+            return update_hilbertspace
+
+        # function is missing self argument; create function with self dummy variable
+        def full_update_func(sweep: "ParameterSweep", *args):
+            return update_hilbertspace(*args)
+
+        return full_update_func
+
     @property
     def bare_specdata_list(self) -> List[SpectrumData]:
         """
@@ -161,7 +195,7 @@ class ParameterSweepBase(ABC):
                     param_vals=self._parameters[sweep_param_name],
                 )
             )
-        self._current_param_indices = None
+        self._current_param_indices = slice(None, None, None)
         return specdata_list
 
     @property
@@ -189,7 +223,7 @@ class ParameterSweepBase(ABC):
             param_name=sweep_param_name,
             param_vals=self._parameters[sweep_param_name],
         )
-        self._current_param_indices = None
+        self._current_param_indices = slice(None, None, None)
         return specdata
 
     def get_sweep_indices(self, multi_index: GIndexTuple) -> List[int]:
@@ -207,7 +241,7 @@ class ParameterSweepBase(ABC):
                 (list, tuple, ndarray),
             )
         ]
-        self._current_param_indices = None
+        self._current_param_indices = slice(None, None, None)
         return sweep_indices
 
     @property
@@ -275,6 +309,7 @@ class ParameterSweepBase(ABC):
         initial: Optional[Union[int, Tuple[int, ...]]] = None,
         final: Optional[Tuple[int, ...]] = None,
         sidebands: bool = False,
+        photon_number: int = 1,
         make_positive: bool = False,
         as_specdata: bool = False,
         param_indices: Optional[NpIndices] = None,
@@ -306,6 +341,9 @@ class ParameterSweepBase(ABC):
         sidebands:
             if set to true, sideband transitions with multiple subsystems changing
             excitation levels are included (default: False)
+        photon_number:
+            number of photons involved in transition; transition frequencies are divided
+            by this number (default: photon_number=1, i.e., single-photon transitions)
         make_positive:
             boolean option relevant if the initial state is an excited state;
             downwards transition energies would regularly be negative, but are
@@ -347,7 +385,7 @@ class ParameterSweepBase(ABC):
             )
 
         if len(initial_state) < len(self._hilbertspace):
-            initial_state = self._complete_initial_state(initial_state, subsys_list)
+            initial_state = self._complete_state(initial_state, subsys_list)
 
         final_states_list = self._get_final_states(
             initial_state, subsys_list, final, sidebands
@@ -355,21 +393,37 @@ class ParameterSweepBase(ABC):
 
         transitions = []
         transition_energies = []
-        initial_energies = self[param_indices].energy_by_bare_index(initial_state)
+
+        if sum(initial_state) == 0:
+            # Identify the (0,0,...,0) state as ground state. Even if it is strongly
+            # hybridized, we can still subtract the true ground state energy. This
+            # addresses issue 103.
+            initial_energies = self[param_indices].energy_by_dressed_index(0)
+        else:
+            initial_energies = self[param_indices].energy_by_bare_index(initial_state)
+
         for final_state in final_states_list:
             final_energies = self[param_indices].energy_by_bare_index(final_state)
             diff_energies = (final_energies - initial_energies).astype(float)
+            diff_energies /= photon_number
             if make_positive:
                 diff_energies = np.abs(diff_energies)
             if not np.isnan(diff_energies.toarray()).all():
                 transitions.append((initial_state, final_state))
                 transition_energies.append(diff_energies)
 
-        self._current_param_indices = None
+        self._current_param_indices = slice(None, None, None)
 
         if not as_specdata:
             return transitions, transition_energies
 
+        label_list = [
+            r"{}$\to${}".format(
+                utils.tuple_to_short_str(elem[0]), utils.tuple_to_short_str(elem[1])
+            )
+            for elem in transitions
+        ]
+
         reduced_parameters = self._parameters.create_sliced(param_indices)
         if len(reduced_parameters) == 1:
             name = reduced_parameters.names[0]
@@ -379,15 +433,16 @@ class ParameterSweepBase(ABC):
                 system_params=self.system_params,
                 param_name=name,
                 param_vals=vals,
-                labels=list(map(str, transitions)),
+                labels=label_list,
                 subtract=np.asarray(
                     [initial_energies] * self._evals_count, dtype=float
                 ).T,
             )
+
         return SpectrumData(
             energy_table=np.asarray(transition_energies),
             system_params=self.system_params,
-            label=list(map(str, transitions)),
+            labels=label_list,
         )
 
     def plot_transitions(
@@ -396,6 +451,7 @@ class ParameterSweepBase(ABC):
         initial: Optional[Union[int, Tuple[int, ...]]] = None,
         final: Optional[Union[int, Tuple[int, ...]]] = None,
         sidebands: bool = False,
+        photon_number: int = 1,
         make_positive: bool = True,
         coloring: Union[str, ndarray] = "transition",
         param_indices: Optional[NpIndices] = None,
@@ -428,6 +484,9 @@ class ParameterSweepBase(ABC):
         sidebands:
             if set to true, sideband transitions with multiple subsystems changing
             excitation levels are included (default: False)
+        photon_number:
+            number of photons involved in transition; transition frequencies are divided
+            by this number (default: photon_number=1, i.e., single-photon transitions)
         make_positive:
             boolean option relevant if the initial state is an excited state;
             downwards transition energies would regularly be negative, but are
@@ -460,20 +519,28 @@ class ParameterSweepBase(ABC):
             initial,
             final,
             sidebands,
+            photon_number,
             make_positive,
             as_specdata=True,
             param_indices=param_indices,
         )
         specdata_all = copy.deepcopy(self[param_indices].dressed_specdata)
         specdata_all.energy_table -= specdata.subtract
+        specdata_all.energy_table /= photon_number
         if make_positive:
             specdata_all.energy_table = np.abs(specdata_all.energy_table)
-        self._current_param_indices = None  # reset from pre-slicing
+        self._current_param_indices = slice(None, None, None)  # reset from pre-slicing
 
         if coloring == "plain":
             return specdata_all.plot_evals_vs_paramvals()
 
-        fig_ax = specdata_all.plot_evals_vs_paramvals(color="gainsboro", linewidth=0.75)
+        if "fig_ax" in kwargs:
+            fig_ax = kwargs.pop("fig_ax")
+        else:
+            fig_ax = None
+        fig_ax = specdata_all.plot_evals_vs_paramvals(
+            color="gainsboro", linewidth=0.75, fig_ax=fig_ax
+        )
 
         labellines_status = plot._LABELLINES_ENABLED
         plot._LABELLINES_ENABLED = False
@@ -483,6 +550,9 @@ class ParameterSweepBase(ABC):
         plot._LABELLINES_ENABLED = labellines_status
         return fig, axes
 
+    def keys(self):
+        return self._data.keys()
+
     def add_sweep(
         self,
         sweep_function: Union[str, Callable],
@@ -517,58 +587,53 @@ class ParameterSweepBase(ABC):
                 )
             sweep_name = sweep_name or sweep_function.__name__
             func = sweep_function
-            self._data[sweep_name] = sweeps.generator(self, func, **kwargs)
+            self._data[sweep_name] = generator(self, func, **kwargs)
         else:
             sweep_name = sweep_name or sweep_function
             func = getattr(sweeps, sweep_function)
             self._data[sweep_name] = func(**kwargs)
 
-    def matrix_elements(
+    def add_matelem_sweep(
         self,
-        operator_name: str,
+        operator: Union[str, Qobj],
         sweep_name: str,
-        subsystem: "QuantumSys",
+        subsystem: "QuantumSys" = None,
     ) -> None:
         """Generate data for matrix elements with respect to a given operator, as a
         function of the sweep parameter(s)
 
         Parameters
         ----------
-        operator_name:
-            name of the operator in question
+        operator:
+            name of the operator in question (str), or full operator in Qobj form
         sweep_name:
             The sweep data will be accessible as <ParameterSweep>[<sweep_name>]
         subsystem:
-            subsystems for which to compute matrix elements.
+            subsystems for which to compute matrix elements, required if operator is
+            given in str form
 
         Returns
         -------
             None; results are saved as <ParameterSweep>[<sweep_name>]
         """
-
-        def _matrix_elements(
-            sweep: "ParameterSweep",
-            param_indices: Tuple[int, ...],
-            param_vals: Tuple[float, ...],
-            operator_name: Union[str, None] = None,
-            subsystem=None,
-        ) -> np.ndarray:
-            subsys_index = sweep.get_subsys_index(subsystem)
-            bare_evecs = sweep["bare_evecs"][subsys_index][param_indices]
-            return subsystem.matrixelement_table(
-                operator=operator_name,
-                evecs=bare_evecs,
-                evals_count=subsystem.truncated_dim,
+        if isinstance(operator, str):
+            operator_func = functools.partial(
+                sweeps.bare_matrixelement,
+                operator_name=operator,
+                subsystem=subsystem,
+            )
+        elif isinstance(operator, Qobj):
+            operator_func = functools.partial(
+                sweeps.dressed_matrixelement,
+                operator=operator,
+            )
+        else:
+            raise TypeError(
+                "Unrecognized type of operator for matrix elements; "
+                "expected: str or Qobj."
             )
 
-        operator_func = functools.partial(
-            _matrix_elements,
-            sweep=self,
-            operator_name=operator_name,
-            subsystem=subsystem,
-        )
-
-        matrix_element_data = sweeps.generator(
+        matrix_element_data = generator(
             self,
             operator_func,
         )
@@ -610,7 +675,12 @@ class ParameterSweep(
         and the set of values to be used in the sweep.
     update_hilbertspace:
         function that updates the associated `hilbertspace` object with a given
-        set of parameters
+        set of parameters; signature is either
+        `update_hilbertspace(paramval1, paramval2, ...)`
+        or
+        `update_hilbertspace(self, paramval1, paramval2, ...)`
+        where `self` makes the `ParameterSweep` instance available, and thereby
+        dict-like access to subsystems and interaction terms
     evals_count:
         number of dressed eigenvalues/eigenstates to keep. (The number of bare
         eigenvalues/eigenstates is determined for each subsystems by `truncated_dim`.)
@@ -635,6 +705,15 @@ class ParameterSweep(
     autorun:
         Determines whether to directly run the sweep or delay it until `.run()` is
         called manually. (Default: `settings.AUTORUN_SWEEP=True`)
+    deepcopy:
+        if set to True, the parameter sweep is run with an exact copy of the Hilbert
+        space; this ensures that all parameters after the sweep are identical to
+        parameters before the sweep. Note: changing global HilbertSpace or
+        QuantumSystem attributes will have no effect with this option; all updates
+        must be made via `<ParameterSweep>.hilbertspace[<id_str>] = ... If
+        set to False (default), updates to global instances have the expected effect.
+        The HilbertSpace object and all its constituent parts are left in the state
+        reached by the very final parameter update.
     num_cpus:
         number of CPU cores requested for computing the sweep
         (default value `settings.NUM_CPUS`)
@@ -649,6 +728,8 @@ class ParameterSweep(
                 "old-style interface will cease to be supported in the future.",
                 FutureWarning,
             )
+            from scqubits.legacy._param_sweep import _ParameterSweep
+
             return _ParameterSweep(*args, **kwargs)
         else:
             return super().__new__(cls, *args, **kwargs)
@@ -662,21 +743,23 @@ class ParameterSweep(
         subsys_update_info: Optional[Dict[str, List[QuantumSys]]] = None,
         bare_only: bool = False,
         autorun: bool = settings.AUTORUN_SWEEP,
+        deepcopy: bool = False,
         num_cpus: Optional[int] = None,
     ) -> None:
         num_cpus = num_cpus or settings.NUM_CPUS
         self._parameters = Parameters(paramvals_by_name)
         self._hilbertspace = hilbertspace
         self._evals_count = evals_count
-        self._update_hilbertspace = update_hilbertspace
+        self._update_hilbertspace = self.set_update_func(update_hilbertspace)
         self._subsys_update_info = subsys_update_info
         self._data: Dict[str, Optional[NamedSlotsNdarray]] = {}
         self._bare_only = bare_only
+        self._deepcopy = deepcopy
         self._num_cpus = num_cpus
         self.tqdm_disabled = settings.PROGRESSBAR_DISABLED or (num_cpus > 1)
 
         self._out_of_sync = False
-        self._current_param_indices = None
+        self._current_param_indices = slice(None, None, None)
 
         dispatch.CENTRAL_DISPATCH.register("PARAMETERSWEEP_UPDATE", self)
         dispatch.CENTRAL_DISPATCH.register("HILBERTSPACE_UPDATE", self)
@@ -686,7 +769,7 @@ class ParameterSweep(
 
     def cause_dispatch(self) -> None:
         initial_parameters = tuple(paramvals[0] for paramvals in self._parameters)
-        self._update_hilbertspace(*initial_parameters)
+        self._update_hilbertspace(self, *initial_parameters)
 
     @classmethod
     def deserialize(cls, iodata: "IOData") -> "StoredSweep":
@@ -719,8 +802,14 @@ class ParameterSweep(
         """Create all sweep data: bare spectral data, dressed spectral data, lookup
         data and custom sweep data."""
         # generate one dispatch before temporarily disabling CENTRAL_DISPATCH
-        self.cause_dispatch()
+
+        if self._deepcopy:
+            stored_hilbertspace = copy.deepcopy(self._hilbertspace)
+            self._hilbertspace = copy.deepcopy(self._hilbertspace)
+        else:
+            self.cause_dispatch()
         settings.DISPATCH_ENABLED = False
+
         self._data["bare_evals"], self._data["bare_evecs"] = self._bare_spectrum_sweep()
         if not self._bare_only:
             self._data["evals"], self._data["evecs"] = self._dressed_spectrum_sweep()
@@ -730,6 +819,8 @@ class ParameterSweep(
                 self._data["chi"],
                 self._data["kerr"],
             ) = self._dispersive_coefficients()
+        if self._deepcopy:
+            self._hilbertspace = stored_hilbertspace  # restore original state
         settings.DISPATCH_ENABLED = True
 
     def _bare_spectrum_sweep(self) -> Tuple[NamedSlotsNdarray, NamedSlotsNdarray]:
@@ -769,7 +860,7 @@ class ParameterSweep(
     def _update_subsys_compute_esys(
         self, update_func: Callable, subsystem: QuantumSys, paramval_tuple: Tuple[float]
     ) -> ndarray:
-        update_func(*paramval_tuple)
+        update_func(self, *paramval_tuple)
         evals, evecs = subsystem.eigensys(evals_count=subsystem.truncated_dim)
         esys_array = np.empty(shape=(2,), dtype=object)
         esys_array[0] = evals
@@ -846,7 +937,7 @@ class ParameterSweep(
         paramindex_tuple: Tuple[int],
     ) -> ndarray:
         paramval_tuple = self._parameters[paramindex_tuple]
-        update_func(*paramval_tuple)
+        update_func(self, *paramval_tuple)
 
         assert self._data is not None
         bare_esys = {
@@ -901,15 +992,18 @@ class ParameterSweep(
                 )
             )
 
-        spectrum_data = np.asarray(spectrum_data, dtype=object)
-        spectrum_data = spectrum_data.reshape((*self._parameters.counts, 2))
+        spectrum_data_ndarray = np.asarray(spectrum_data, dtype=object)
+        spectrum_data_ndarray = spectrum_data_ndarray.reshape(
+            (*self._parameters.counts, 2)
+        )
         slotparamvals_by_name = OrderedDict(self._parameters.ordered_dict.copy())
 
-        evals = np.asarray(spectrum_data[..., 0].tolist())
-        evecs = spectrum_data[..., 1]
+        evals = np.asarray(spectrum_data_ndarray[..., 0].tolist())
+        evecs = spectrum_data_ndarray[..., 1]
 
-        return NamedSlotsNdarray(evals, slotparamvals_by_name), NamedSlotsNdarray(
-            evecs, slotparamvals_by_name
+        return (
+            NamedSlotsNdarray(evals, slotparamvals_by_name),
+            NamedSlotsNdarray(evecs, slotparamvals_by_name),
         )
 
     def _energies_1(self, subsys):
@@ -1011,19 +1105,25 @@ class StoredSweep(
     dispatch.DispatchClient,
     serializers.Serializable,
 ):
-    _parameters = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
-    _evals_count = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
-    _data = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _parameters: Parameters = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _evals_count: int = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
+    _data: Dict[str, Any] = descriptors.WatchedProperty("PARAMETERSWEEP_UPDATE")
     _hilbertspace: HilbertSpace
 
-    def __init__(self, paramvals_by_name, hilbertspace, evals_count, _data) -> None:
+    def __init__(
+        self,
+        paramvals_by_name: Dict[str, Union[ndarray, Iterable]],
+        hilbertspace: HilbertSpace,
+        evals_count: int,
+        _data,
+    ) -> None:
         self._parameters = Parameters(paramvals_by_name)
         self._hilbertspace = hilbertspace
         self._evals_count = evals_count
         self._data = _data
 
         self._out_of_sync = False
-        self._current_param_indices = None
+        self._current_param_indices: NpIndices = slice(None, None, None)
 
     @classmethod
     def deserialize(cls, iodata: "IOData") -> "StoredSweep":
@@ -1066,3 +1166,69 @@ class StoredSweep(
             autorun=autorun,
             num_cpus=num_cpus,
         )
+
+
+def generator(sweep: "ParameterSweepBase", func: Callable, **kwargs) -> np.ndarray:
+    """Method for computing custom data as a function of the external parameter,
+    calculated via the function `func`.
+
+    Parameters
+    ----------
+    sweep:
+        ParameterSweep object containing HilbertSpace and spectral information
+    func:
+        signature: `func(parametersweep, paramindex_tuple, paramvals_tuple,
+        **kwargs)`, specifies how to calculate the data for a single choice of
+        parameter(s)
+    **kwargs:
+        keyword arguments to be included in func
+
+    Returns
+    -------
+        array of custom data
+    """
+    # obtain reduced parameters from pre-slicing info
+    reduced_parameters = sweep._parameters.create_sliced(
+        sweep._current_param_indices, remove_fixed=False
+    )
+    total_count = np.prod(reduced_parameters.counts)
+
+    def func_effective(paramindex_tuple: Tuple[int], params, **kw) -> Any:
+        paramvals_tuple = params[paramindex_tuple]
+        return func(
+            sweep,
+            paramindex_tuple=paramindex_tuple,
+            paramvals_tuple=paramvals_tuple,
+            **kw,
+        )
+
+    if hasattr(func, "__name__"):
+        func_name = func.__name__
+    else:
+        func_name = ""
+
+    data_array = list(
+        tqdm(
+            map(
+                functools.partial(
+                    func_effective,
+                    params=reduced_parameters,
+                    **kwargs,
+                ),
+                itertools.product(*reduced_parameters.ranges),
+            ),
+            total=total_count,
+            desc="sweeping " + func_name,
+            leave=False,
+            disable=settings.PROGRESSBAR_DISABLED,
+        )
+    )
+    element_shape: Tuple[int, ...] = tuple()
+    if isinstance(data_array[0], np.ndarray):
+        element_shape = data_array[0].shape
+
+    data_ndarray = np.asarray(data_array)
+    return NamedSlotsNdarray(
+        data_ndarray.reshape(reduced_parameters.counts + element_shape),
+        reduced_parameters.paramvals_by_name,
+    )
